-- |Ingress Graph module.
-- The routines in this module support optimizing links between
-- portals in the game Ingress.
module InGraph where

import Data.List
import qualified Data.Set as Set
import Data.Maybe
import System.Random
import Data.Graph.Inductive

data Vector a = Vec (a,a) deriving (Eq, Show)
type Portal = (String,Vector Float)
type PortalGraph = Gr Portal ()

emptyGraph = empty::PortalGraph

-- Sample portal list, used for testing
portals = [
--    ("Montana Ave Library", Vec (17.1,10.0)),
--    ("Douglas park", Vec (25.0,6.1)),
--    ("Mural (new balance)", Vec (26.1,6.0)),
--    ("Artistic house", Vec (26.0,7.0)),
--    ("St. Monica (church)", Vec (7.1,7.1)),
--    ("Post Office (26th St)", Vec (26.0,14.1)),
--    ("Museum Systems", Vec (27.0,14.0))
    ("Montana Ave Library", Vec (430,232)),
    ("Douglas park", Vec (588,221)),
    ("Mural (new balance)", Vec (605,213)),
    ("Artistic house", Vec (581,191)),
    ("St. Monica (church)", Vec (372,395)),
    ("Post Office (26th St)", Vec (447,51)),
    ("Museum Systems", Vec (461,41)),
    ("Palisades Branch Library", Vec (46,102)),
    ("Fire Station 69", Vec (77,86)),
    ("Post Office (Palisades)", Vec (40,72))
    ]

northSaMo = addPortals emptyGraph portals

-- |Add a list of portals to a portal graph
addPortals :: PortalGraph -> [Portal] -> PortalGraph
addPortals g [] = g
addPortals g (p:ps) = addPortals (insNode (newNode g, p) g) ps

-- |Determine the id of the next node to add
newNode :: PortalGraph -> Node
newNode g = (newNodes 1 g)!!0

mean x = (sum x) / (fromIntegral $ length x)
std x = (foldl (\b c -> b + c^2) 0 (map (\a -> a - (mean x)) x)) / (fromIntegral $ length x - 1)

-- |Determine whether or not links between two pairs of portals
-- cross
linksCross :: (Portal,Portal) -> (Portal,Portal) -> Bool
linksCross link1 link2 =
    let p1 = snd (fst link1)
        p2 = snd (snd link1)
        p3 = snd (fst link2)
        p4 = snd (snd link2)
        v = p2 - p1
        w = p4 - p3
        denom = cross2D v w
    in if denom /= 0
       then let crossPointv = (cross2D (p3 - p1) w) / denom
                crossPointw = (cross2D (p3 - p1) v) / denom
            in (crossPointw > 0.0 && crossPointw < 1.0) && (crossPointv > 0.0 && crossPointv < 1.0)
       else False

-- |Two dimensional cross product, used for link crossing
cross2D :: Num a => Vector a -> Vector a -> a
cross2D (Vec (v1,v2)) (Vec (w1,w2)) = v1*w2 - v2*w1

-- |Determine the energy of a graph
hamiltonian :: Float -> PortalGraph -> Float
hamiltonian gain g = let energies = map (portalEnergy gain g) (nodes g)
                in (sum energies)

-- |Count the number of fields adjacent to a portal.
-- Returns a list of endpoint pairs for each triangle.
-- For instance:
--
-- >>> adjFields g 1
-- [(2,3),(3,4)]
-- 
-- Specifies that node 1 makes triangles (1,2,3) and (1,3,4).
adjFields :: PortalGraph -> Node -> [(Node,Node)]
adjFields g p = let links = suc (undir g) p
                    endpoints = map (intersect links) (map (suc g) links)
                    endpairs = map (\a -> zip (repeat (fst a)) (snd a) ) (zip links endpoints)
                in concat endpairs
-- |Calculate the area covered by a field
fieldArea :: Portal -> Portal -> Portal -> Float
fieldArea portal1 portal2 portal3 =
    let p1 = snd portal1
        p2 = snd portal2
        p3 = snd portal3
        a = norm (p2-p1)
        b = norm (p3-p2)
        c = norm (p1-p3)
        s = (a+b+c)/2
    in sqrt (s*(s-a)*(s-b)*(s-c))
    where norm (Vec (x,y)) = sqrt (x^2+y^2)

-- |Calculate the area covered by a field,
-- using graph node numbers
fieldAreaG :: PortalGraph -> Node -> Node -> Node -> Float
fieldAreaG g p1 p2 p3 =
    let portal1 = fromJust $ lab g p1
        portal2 = fromJust $ lab g p2
        portal3 = fromJust $ lab g p3
    in fieldArea portal1 portal2 portal3

-- |Calculate the energy of a single portal.
-- Energy is defined to be the ratio of AP
-- due to destruction over AP due to creation.
-- This is then scaled by the number of fields.
portalEnergy :: Float -> PortalGraph -> Node -> Float
portalEnergy fieldGain g p =
    let links = deg g p
        fields = adjFields g p
        numFields = length fields
        --area = foldl (\a f -> a + (fieldAreaG g p (fst f) (snd f))) 0 fields
        enemyAP = fromIntegral (187*links + 750*numFields)
        friendlyAP = fromIntegral (313*links + 1250*numFields)
    in if links < 1
          then 1e6
          else enemyAP / friendlyAP / (1-fieldGain+fieldGain*(tanh $ fromIntegral (1+numFields)))

-- |Determine whether a potential link would cross
-- an existing link
makesCross :: PortalGraph -> (Node,Node) -> Bool
makesCross g (from,to) =
    let fromNode = fromJust $ lab g from
        toNode = fromJust $ lab g to
        crosses = map (\(a,b)->linksCross (fromNode,toNode) (fromJust $ lab g a, fromJust $ lab g b)) (edges g)
    in or crosses

-- |Implement a simple metropolis algorithm for minimizing
-- the hamiltonian of a portal graph
metropolis :: StdGen -> PortalGraph -> Float -> Int -> PortalGraph
metropolis randgen g gain 0 = g
metropolis randgen g gain n =
    let (temp,newRandgen) = perturbGraph randgen g
    in if (hamiltonian gain temp) < (hamiltonian gain g)
          then metropolis newRandgen temp gain (n-1)
          else metropolis newRandgen g gain (n-1)

-- |Create a random modification to a graph.
-- Pick a link to toggle on or off.
-- Continue toggling links until one of the toggles adds a link.
-- This is because the hamiltonian will nearly always judge removal of a link as a higher energy state.
perturbGraph :: StdGen -> PortalGraph -> (PortalGraph,StdGen)
perturbGraph randgen g = 
    -- Select two random nodes for a link
    let (randLinkFrom,randgen2) =  randomR (1, noNodes g) randgen
        (randLinkTo,randgen3) =  randomR (1, noNodes g) randgen2
    --- If it's a self-link, try again
    in if randLinkFrom == randLinkTo
       then if (length $ nodes g) < 2
            then (g,randgen3)
            else perturbGraph randgen3 g
       else 
          --Toggle links until a toggle adds a link
          if (randLinkFrom,randLinkTo) `elem` (edges g)
             then perturbGraph randgen3 (delEdge (randLinkFrom,randLinkTo) g)
             else if (randLinkTo,randLinkFrom) `elem` (edges g)
                  then perturbGraph randgen3 (delEdge (randLinkTo,randLinkFrom) g)
                  else if makesCross g (randLinkFrom,randLinkTo)
                       then perturbGraph randgen3 g
                       else (insEdge (randLinkFrom,randLinkTo,()) g, randgen3)
 
-- |Remove all of the edges from a graph
stripEdges :: PortalGraph -> PortalGraph
stripEdges g =
    delEdges (edges g) g

-- |Optimize the links among a collection of nodes.
graphOptimize :: Float -> Int -> PortalGraph -> PortalGraph
graphOptimize gain maxiter portalGraph =
    metropolis (mkStdGen 1) portalGraph gain maxiter

-- |Calculate some descriptive numbers summarizing a graph
graphStats :: PortalGraph -> (Int, Int, Int)
graphStats g =
    let links = sum $ map (outdeg g) (nodes g)
        allAdjFields = map (adjFields g) (nodes g)
        fieldVerts = zipWith (\a b -> map (\c-> [a, (fst c), (snd c)]) b)  (nodes g) allAdjFields
        uniqueFields = nub $ map Set.fromList $ concat fieldVerts
        numFields = length uniqueFields
        enemyAP = 187*links + 750*numFields
        friendlyAP = 313*links + 1250*numFields
    in (enemyAP,friendlyAP,numFields)
     

-- Class instances to help dealing with the Vector type
instance Functor Vector where
    fmap f (Vec (x,y)) = Vec (f x, f y)

-- | Allow math to be used with a 'Vector'
instance Num a => Num (Vector a) where
    (+) (Vec (a1,a2)) (Vec (b1,b2)) = Vec (a1+b1, a2+b2)
    (-) (Vec (a1,a2)) (Vec (b1,b2)) = Vec (a1-b1, a2-b2)
    (*) (Vec (a1,a2)) (Vec (b1,b2)) = Vec (a1*b1, a2*b2)
    negate (Vec (a1,a2)) = Vec (-a1,-a2)
    abs (Vec (a1,a2)) = Vec (abs a1, abs a2)
    signum (Vec (a1,a2)) = Vec (signum a1, signum a2)
    fromInteger x = Vec (fromInteger x, fromInteger x)

