-- |InGraph, an ingress graph optimizer
module Main where

import InGraph
import DataDict
import Data.Maybe
import Data.Graph.Inductive
import Graphics.UI.WX
import Control.Exception



main :: IO ()
main = start doFrame



-- Specify image files for the file open dialog.
imageFiles
   = [("Image files",["*.bmp","*.jpg","*.gif","*.png"])
     ,("Portable Network Graphics (*.png)",["*.png"])
     ,("BMP files (*.bmp)",["*.bmp"])
     ,("JPG files (*.jpg)",["*.jpg"])
     ,("GIF files (*.gif)",["*.gif"])
     ]


-- | GUI entry point. Set up window elements and define callbacks
doFrame :: IO ()
doFrame
    = do f <- frame [ text := "InGraph: Ingress Link Optimizer" ]

         -- Mutable variables
         portalMap <- variable [ value := Nothing ]
         portalGraph <- variable [ value := emptyGraph ]

         -- Window where we draw the portal graph
         w <- scrolledWindow f [ scrollRate := sz 10 10,
                                 bgcolor := white,
                                 fullRepaintOnResize := False ]

         -- Menu items
         file <- menuPane [ text := "&File" ]
         open <- menuItem file [ text := "&Open" ]
         mclose <- menuItem file [ text := "&Close" ]
         menuLine file
         quit <- menuQuit file [help := "Quit"]


         status <- statusField [ text := "Open a portal map image to begin" ]

         -- Panel for controls and info text
         p <- panel f []
         gainSlider <- hslider p True 0 100 [ selection := 2 ]

         eapText <- staticText p [ text := "      " ]
         fapText <- staticText p [ text := "      " ]
         ratioText <- staticText p [ text := "     " ]
         fieldText <- staticText p [ text := "   " ]

         let textControls = (eapText, fapText, ratioText, fieldText)

         optButton <- button p [ text := "Optimize!", enabled := False ]
         resetButton <- button p [ text := "Reset" ]

         -- Store items in a data dictionary
         let dataDict = DataDict {
                buttonDict = Dict [(ButtonKey OptButton, optButton),
                                   (ButtonKey ResetButton, resetButton)],
                textDict = Dict [(TextKey EAPText, eapText),
                                 (TextKey FAPText, fapText),
                                 (TextKey RatioText, ratioText),
                                 (TextKey FieldText, fieldText)],
                controlDict = Dict [(ControlKey GainSlider, gainSlider)]
                }

         set w [ on click := addPortal f w portalGraph dataDict,
                 on paint := onPaint portalMap portalGraph dataDict ]

         set optButton [ on command := optimizeLinks w portalGraph dataDict ]
         set resetButton [ on command := clearGraph w portalGraph dataDict ]

         -- Lay out the frame using the bits defined above
         set f [ layout := fill $ row 5 [
                               fill (widget w),
                               vfill $ boxed "Tools" $ container p $ column 5 [
                                   row 2 [ widget optButton, widget resetButton ],
                                   label "Field Gain (%):",
                                   hfill $ widget gainSlider,
                                   row 2 [label "Enemy AP: ", hfill $ widget eapText],
                                   row 2 [label "Friendly AP: ", hfill $ widget fapText],
                                   label "Ratio (lower is better):",
                                   hfill $ widget ratioText,
                                   row 2 [label "Fields: ",
                                          hfill $ widget fieldText]
                                   ]
                               ],
                 statusBar := [status],
                 menuBar := [file],
                 outerSize := sz 400 300,
                 on (menu quit) := close f,
                 on (menu open) := onOpen f w portalMap portalGraph mclose status,
                 on (menu mclose) := onClose w portalMap mclose status
                 ]


    where
        onOpen :: Frame a -> ScrolledWindow b -> Var (Maybe (Bitmap ())) -> Var PortalGraph -> MenuItem c -> StatusField -> IO ()
        onOpen f w portalMap portalGraph mclose status
            = do imageFile <- fileOpenDialog f False True "Open image" imageFiles "" ""
                 case imageFile of
                    Nothing -> return ()
                    Just file -> openImage w portalMap mclose status file
                 varSet portalGraph emptyGraph

        onClose w portalMap mclose status
            = do repaint w

        openImage w portalMap mclose status file
            = do bm <- bitmapCreateFromFile file
                 set portalMap [ value := Just bm ]
                 set status [ text := file ]
                 bmsize <- get bm size
                 set w [ virtualSize := bmsize ]
                 repaint w
              `onException` repaint w

        onPaint :: Var (Maybe (Bitmap ())) -> Var PortalGraph -> DataDict a b c -> DC d -> Rect -> IO ()
        onPaint portalMap portalGraph dataDict dc viewArea
            = do bmap <- get portalMap value
                 case bmap of
                    Nothing -> return ()
                    Just bm -> drawBitmap dc bm pointZero False []
                 drawPortalGraph dc portalGraph

        -- |Given a graph, draw the nodes and edges
        drawPortalGraph :: DC a -> Var PortalGraph -> IO ()
        drawPortalGraph dc portalGraph = do
            g <- get portalGraph value
            let portals = map snd (labNodes g)
                links = map (mapTuple (fromJust.lab g)) (edges g)
            drawPortalLinks dc (map (mapTuple (mapSnd vec2point)) links)
            drawPortals dc (map (mapSnd vec2point) portals)

        -- |Draw links between portals
        drawPortalLinks :: DC a -> [((String,Point),(String,Point))] -> IO ()
        drawPortalLinks dc [] = return ()
        drawPortalLinks dc (l:ls) = do
            let portal1 = fst l
                portal2 = snd l
                pos1 = snd portal1
                pos2 = snd portal2
            set dc [pen := penColored red 3]
            line dc pos1 pos2 []
            drawPortalLinks dc ls

        -- |Draw a dot for each portal in the graph
        drawPortals :: DC a -> [(String,Point)] -> IO ()
        drawPortals dc [] = return ()
        drawPortals dc (p:ps)
            = do
                let portalName = fst p
                    portalPos = snd p
                set dc [brushColor := red, brushKind := BrushSolid]
                circle dc portalPos 5 []
                drawText dc portalName portalPos [textColor := red, fontFamily := FontSwiss]
                drawPortals dc ps

        -- |Given a click location, add a new portal to the graph
        addPortal :: (Able c) => Frame a -> ScrolledWindow b -> Var PortalGraph -> DataDict c d e -> Point -> IO ()
        addPortal f w portalGraph dataDict pt
            = do portalName <- textDialog f "Please enter a name for this portal" "Name" ""
                 let optButton = getButton OptButton dataDict
                 if (length portalName) > 0
                     then do set optButton [ enabled := True ]
                             varUpdate portalGraph (\g -> addPortals g [(portalName,point2vec pt)] )
                     else varGet portalGraph
                 repaint w

        -- |Utility function to map a function to both elements
        -- of a tuple
        mapTuple :: (a -> b) -> ( (a,a) -> (b,b) )
        mapTuple f = ( f >< f )

        -- |Convert a wxHaskell Point to a Vector used by the
        -- graph routines
        point2vec :: Point -> InGraph.Vector Float
        point2vec (Point x y) = Vec (fromIntegral x, fromIntegral y)

        -- |Convert a Vector to a wxHaskell Point
        vec2point :: InGraph.Vector Float -> Point
        vec2point (Vec (x,y)) = Point (truncate x) (truncate y)

        clearGraph w portalGraph dataDict
            = do let optButton = getButton OptButton dataDict
                 varSet portalGraph emptyGraph
                 set (getText EAPText dataDict) [ text := "" ]
                 set (getText FAPText dataDict) [ text := "" ]
                 set (getText RatioText dataDict) [ text := "" ]
                 set (getText FieldText dataDict) [ text := "" ]
                 set optButton [ enabled := False ]
                 repaint w
{-
        clearGraph w portalGraph textControls optButton
            = do varSet portalGraph emptyGraph
                 let (eapText,fapText,ratioText,fieldText) = textControls
                 set eapText [ text := "" ]
                 set fapText [ text := "" ]
                 set fieldText [ text := "" ]
                 set ratioText [ text := "" ]
                 set optButton [ enabled := False ]
                 repaint w
-}                 

        -- | Call "InGraph" optimization routine on the currently
        -- defined graph
        optimizeLinks w portalGraph dataDict
            = do let gainSlider = getControl GainSlider dataDict
                     eapText = getText EAPText dataDict
                     fapText = getText FAPText dataDict
                     ratioText = getText RatioText dataDict
                     fieldText = getText FieldText dataDict
                 fieldGain <- get gainSlider selection
                 varUpdate portalGraph stripEdges
                 varUpdate portalGraph (graphOptimize ((fromIntegral fieldGain)/100.0) 100)
                 optimizedGraph <- varGet portalGraph
                 let (eap,fap,fields) = graphStats optimizedGraph
                 set eapText [ text := (show eap) ]
                 set fapText [ text := (show fap) ]
                 set fieldText [ text := (show fields) ]
                 set ratioText [ text := (show ((fromIntegral eap)/(fromIntegral fap))) ]
                 repaint w



















